#include <stdio.h>
#include <math.h>
#include <string.h>
#include "ppm.h"

double red_scale = 0.3;
double green_scale = 0.59;
double blue_scale = 0.11;

typedef struct
{
	double r, g, b;
	char code[8];
} color;

void rgb_to_hls(double r, double g, double b, double *h, double *l, double *s)
{
	double max;
	double min;
	double diff;
	double r_dist;
	double g_dist;
	double b_dist;

	max = r;
	if (max < g) max = g;
	if (max < b) max = b;

	min = r;
	if (min > g) min = g;
	if (min > b) min = b;

	diff = max - min;
	*l = (max + min) / 2;

	if (fabs(diff) < 0.00001)
	{
		*s = 0;
		*h = 0;	/* h is undefined */
	}
	else
	{
		if (*l <= 0.5)
			*s = diff / (max + min);
		else
			*s = diff / (2.0 - max - min);

		r_dist = (max - r) / diff;
		g_dist = (max - g) / diff;
		b_dist = (max - b) / diff;

		if (r == max)
			*h = b_dist - g_dist;
		else if (g == max)
			*h = 2.0 + r_dist - b_dist;
		else
			*h = 4.0 + g_dist - r_dist;

		*h *= 60.0;
		if (*h < 0.0) *h += 360.0;
	}
}

void version(void)
{
	printf("ppmtoansi v" VERSION ", (C) 2007-2010 by folkert@vanheusden.com\n\n");
}

void help(void)
{
	version();

	printf("-8    use also ascii characters > 126\n");
	printf("-c x  r: remap control characters, s: replace by spaces\n");
	printf("-R x  red scale (default: %f)\n", red_scale);
	printf("-G x  green scale (default: %f)\n", green_scale);
	printf("-B x  blue scale (default: %f)\n", blue_scale);
	printf("-m x  mode: default is rgb, alternative is hls\n");
	printf("-a    ascii only\n");
	printf("-h    this help\n");
	printf("-V    show version & exit\n");
}

typedef enum { MODE_RGB = 0, MODE_HLS } cmode_t;

int main(int argc, char *argv[])
{
	char *filename = NULL;
	FILE *fh = NULL;
	pixel **image;
	int width = 0, height = 0;
	pixval maxval = 0;
	int y;
	char allow_8bit = 0;
	char control_bytes = 'r';	/* r-emap, s-paces */
	int c;
	char ansi = 1;
	int last_color_selected = -1;
	cmode_t mode = MODE_HLS;
	color colors[16] = {
		{ 0,   0,   0,   "[0;30m" },	/* black   */
		{ 0.5, 0,   0,   "[0;31m" },	/* red     */
		{ 0,   0.5, 0,   "[0;32m" },	/* green   */
		{ 0.5, 0.5, 0,   "[0;33m" },	/* yellow  */
		{ 0,   0,   0.5, "[0;34m" },	/* blue    */
		{ 0.5, 0,   0.5, "[0;35m" },	/* magenta */
		{ 0,   0.5, 0.5, "[0;36m" },	/* cyan    */
		{ 0.5, 0.5, 0.5, "[0;37m" },	/* gray    */
		{ 0.25, 0.25, 0.25, "[1;30m" },	/* dark gray */
		{ 1,   0,   0,   "[1;31m" },	/* red     */
		{ 0,   1,   0,   "[1;32m" },	/* green   */
		{ 1,   1,   0,   "[1;33m" },	/* yellow  */
		{ 0,   0,   1,   "[1;34m" },	/* blue    */
		{ 1,   0,   1,   "[1;35m" },	/* magenta */
		{ 0,   1,   1,   "[1;36m" },	/* cyan    */
		{ 1,   1,   1,   "[1;37m" },	/* gray    */
	};
	while((c = getopt(argc, argv, "ahV8c:R:G:B:m:")) != -1)
	{
		switch(c)
		{
			case 'a':
				ansi = 0;
				break;

			case 'm':
				if (strcasecmp(optarg, "hls") == 0)
					mode = MODE_HLS;
				else if (strcasecmp(optarg, "rgb") == 0)
					mode = MODE_RGB;
				else
				{
					fprintf(stderr, "-m requires either 'rgb' or 'hls' as parameter\n");
					return 1;
				}
				break;

			case 'R':
				red_scale = atof(optarg);
				break;

			case 'G':
				green_scale = atof(optarg);
				break;

			case 'B':
				blue_scale = atof(optarg);
				break;

			case '8':
				allow_8bit = 1;
				break;

			case 'c':
				control_bytes = optarg[0];
				if (control_bytes != 'c' && control_bytes != 's')
				{
					help();
					return 1;
				}
				break;
			case 'h':
				help();
				return 0;

			case 'V':
				version();
				return 0;

			default:
				help();
				return 1;
		}
	}



	if (filename == NULL)
	{
		fh = fdopen(0, "r");
	}
	else
	{
		fh = fopen(filename, "r");
	}

	if (!fh)
	{
		fprintf(stderr, "Failed to open file %s.\n", filename);
		return 1;
	}

	image = ppm_readppm(fh, &width, &height, &maxval); 
	if (!image)
	{
		fprintf(stderr, "Failed to read file.\n");
		return 1;
	}

	for(y=0; y<height; y++)
	{
		int x;

		for(x=0; x<width; x++)
		{
			double r = (double)PPM_GETR(image[y][x]) / (double)maxval;
			double g = (double)PPM_GETG(image[y][x]) / (double)maxval;
			double b = (double)PPM_GETB(image[y][x]) / (double)maxval;
			double hc, lc, sc;
			double gray = '.';
			int max_c;
			unsigned char c;
			int loop, color_index = 15;
			double min_color_diff = 65535;

			/* determine character to display */
			if (mode == MODE_RGB)
			{
				gray = red_scale * r + green_scale * g + blue_scale * b;
			}
			else if (mode == MODE_HLS)
			{
				rgb_to_hls(r, g, b, &hc, &lc, &sc);
				gray = lc;
			}

			if (allow_8bit)
				max_c = 255;
			else
				max_c = 127;

			if (control_bytes == 'r')
				max_c -= 33;

			c = gray * max_c;

			if (control_bytes == 'r')
			{
				c += 32;

				if (c > 126)
				{
					if (allow_8bit)
						c++;
					else
						c = ' ';
				}
			}
			else if (control_bytes == 's')
			{
				if (c < 32)
					c = ' ';
				else if (c == 127)
					c = ' ';
			}

			/* find color */
			for(loop=0; loop<16; loop++)
			{
				double cur_diff = 0.0;

				switch(mode)
				{
					case MODE_RGB:
						cur_diff = fabs(colors[loop].r - r) * red_scale +
							   fabs(colors[loop].g - g) * green_scale +
							   fabs(colors[loop].b - b) * blue_scale;
						break;

					case MODE_HLS:
						{
							double hi, li, si;

							rgb_to_hls(colors[loop].r, colors[loop].g, colors[loop].b, &hi, &li, &si);

							cur_diff = fabs(hi - hc) +
								   fabs(li - lc) +
								   fabs(si - sc);
						}
						break;
				}

				if (cur_diff < min_color_diff)
				{
					min_color_diff = cur_diff;
					color_index = loop;

					if (cur_diff == 0.0)
						break;
				}
			}

			if (color_index != last_color_selected)
			{
				if (ansi)
					printf("%c%s%c", 27, colors[color_index].code, c);
				else
					printf("%c", c);
				last_color_selected = color_index;
			}
			else
				printf("%c", c);
		}

		printf("\n");
	}

	if (ansi)
		printf("%c%s\n", 27, colors[15].code);

	return 0;
}
